#!/usr/bin/env python
'''Node for estimating environmental forces
Author: Maiken Berthelsen
'''
from std_msgs.msg import String, Int64, Float64
import rospy
import numpy as np
import models
from custom_msgs.msg import NorthEastHeading, ThreeDofForce, SurgeSwayYaw
import custom_msgs.msg
import custom_msgs.topics as topics
import std_msgs.msg
import csv
from csv import writer
import math
from math import radians, pi
from scipy import linalg

class Kalman:
	def __init__(self, x, P):
		self.not_first = False
		self.x_prev = x
		self.P_prev = P
		self.x_est = self.x_prev  
		self.P_est = self.P_prev
		self.x_pred = self.x_prev 
		self.P_pred = self.P_prev

		#load the milliampere parameters
		params = rospy.get_param('~env_estimation/milliampere')
		self.milliampere = models.MilliampereModel(params)
		self.M_inv = self.milliampere.M_inv.astype(float)
		self.M_inv_N = None
		self.yaw_ang_est = 0

		# define the gains
		Q_array = np.array([1.0,1.0,0.01,1.0,1.0,0.01,10,10,0.0,1.0*(10**2),1.0*(10**2),2.0*(10**2)])
		R_array = np.array([3.0*(10**-2),3.0*(10**-2),9.0*(10**-4),4.0*(10**-2),4.0*(10**-2),1.0*(10**-3)])

		self.Q = np.diag(Q_array)		
		self.cov_R = np.diag(R_array)

		self.E = np.vstack([np.zeros((3,12)),\
			np.hstack([np.zeros((3,3)), self.M_inv, np.zeros((3,3)), np.zeros((3,3))]),\
			np.hstack([np.zeros((2,6)), np.eye(2), np.zeros((2,1)),np.zeros((2,3))]),\
			np.zeros((1,12)), np.hstack([np.zeros((3,9)), np.eye(3)])])
		
		self.H_mat = np.vstack([np.hstack([np.eye(3), np.zeros((3,3)), np.zeros((3,3)),np.zeros((3,3))]),\
			np.hstack([np.zeros((3,3)),np.eye(3),np.zeros((3,3)),np.zeros((3,3))])])

		self.N = None
		self.y = np.zeros((6,1))
		self.T_thrust = 1.0;		
		self.kalman_gain_pub = rospy.Publisher('KalmanGain', Float64, queue_size=10)
		self.kalman_pub_eta = rospy.Publisher("/KalmanEta", NorthEastHeading, queue_size=10)
		self.kalman_pub_nu = rospy.Publisher("/KalmanNu", SurgeSwayYaw, queue_size=10)
		self.kalman_pub_nu_c = rospy.Publisher("/KalmanNuCurrent", ThreeDofForce, queue_size=10)

		#debuggging
		self.kalman_pub_tau = rospy.Publisher("/KalmanTau", ThreeDofForce, queue_size=10)

		self.filter_eta_pub = rospy.Publisher("/FilteredEta", NorthEastHeading, queue_size=10)
		self.filter_nu_pub = rospy.Publisher("/FilteredNu", SurgeSwayYaw, queue_size=10)


		

	@staticmethod
	def wrap_pi(angle):
		return (angle + np.pi) % (2*np.pi) - np.pi

	def wrap_yaw(self,psi):
		out = np.remainder(psi,2*pi)			 
		if out > pi:
			out = out - 2*pi
		elif out < -pi:
			out = out + 2*pi
		return out

	def publish_message(self):
		eta_pred = NorthEastHeading(self.x_pred[0,0],self.x_pred[1,0],self.x_pred[2,0])
		nu_pred = SurgeSwayYaw(self.x_pred[3,0],self.x_pred[4,0],self.x_pred[5,0])
		nu_c_pred = ThreeDofForce(self.x_pred[6,0],self.x_pred[7,0],self.x_pred[8,0])
		tau_pred = ThreeDofForce(self.x_pred[9,0],self.x_pred[10,0],self.x_pred[11,0])
		rospy.logwarn('current %.5f %.5f', self.x_pred[6,0],self.x_pred[7,0])
		eta_test = NorthEastHeading(self.y[0,0],self.y[1,0],self.y[2,0])
		nu_test = SurgeSwayYaw(self.y[3,0],self.y[4,0],self.y[5,0])
		K_norm = np.linalg.norm(self.K.astype(float), 2)

		self.kalman_gain_pub.publish(K_norm)
		self.kalman_pub_eta.publish(eta_pred)
		self.kalman_pub_nu.publish(nu_pred)
		self.kalman_pub_nu_c.publish(nu_c_pred)

		self.kalman_pub_tau.publish(tau_pred)
		self.filter_eta_pub.publish(eta_test)
		self.filter_nu_pub.publish(nu_test)



	def rotMatrix(self,psi):
		return np.matrix([[np.cos(psi), -np.sin(psi), 0], [np.sin(psi), np.cos(psi), 0],[0, 0, 1]])

	def shortest_angle_path(self,start, end):
		diff = (end - start + pi)%(2*pi) - pi
		if diff < -pi:
			move = diff + 2*pi
		else:
			move = diff
		return move

	def kalmanGain(self):
		
		P_H_T = np.matmul(self.P_prev,np.transpose(self.H_mat))
		HPH_R=np.matmul(self.H_mat,P_H_T)+self.cov_R
		inv_HPH_R = np.linalg.inv(HPH_R.astype(float))
		return np.matmul(P_H_T,inv_HPH_R)

	def estimate(self,meas):
		self.K = self.kalmanGain()
		self.y[0:3,0] = meas.eta
		self.y[3:6,0] = meas.nu
		self.yaw_ang_est = self.y[2,0]#meas.heading_des
		H_x = self.H_mat.dot(self.x_prev)
		
		if self.not_first:
			self.error = self.y-H_x
			angle = self.shortest_angle_path(H_x[2,0],self.y[2,0])
			self.error[2,0] = angle 
			K_Y_Hx = self.K.dot(self.error)
			self.x_est = self.x_prev + K_Y_Hx
			self.x_est[2,0] = self.wrap_yaw(self.x_est[2,0])
			I_KH = np.eye(12) - np.matmul(self.K,self.H_mat)
			KRK = np.matmul(self.K,np.matmul(self.cov_R,np.transpose(self.K)))
			self.P_est = np.matmul(I_KH,np.matmul(self.P_prev,np.transpose(I_KH))) + KRK
		else:
			self.not_first = True


	def updatePhi(self,h):
		# |    |    |    |    |
		# | 11 | 12 | 13 | 14 |
		# |____|____|____|____|
		# |    |    |    |    |
		# | 21 | 22 | 23 | 24 |
		# |____|____|____|____|
		# |    |    |    |    |
		# | 31 | 32 | 33 | 34 |
		# |____|____|____|____|
		# |    |    |    |    |
		# | 41 | 42 | 43 | 44 |
		# |    |    |    |    |
		# Psi = I + h*Jacobi(f(x,u))|x=x_est

		R = self.rotMatrix(self.yaw_ang_est)
		# derivated rotation matrix wrt psi
		derivated_rot = np.matrix([[-np.sin(self.yaw_ang_est), -np.cos(self.yaw_ang_est), 0],\
			[np.cos(self.yaw_ang_est), -np.sin(self.yaw_ang_est), 0], \
			[0, 0, 0]])

		nu_arr = np.array([[self.x_est[3,0]],[self.x_est[4,0]],[self.x_est[5,0]]]).reshape(3,1)
		partial_jacobi_11 = derivated_rot.dot(nu_arr)

		jacobi_11 = np.hstack([np.zeros((3,2)), partial_jacobi_11])
		jacobi_12 = R

		# estimated current velocity in NED frame
		nu_c_temp = np.zeros((3,1))
		nu_c_temp[0] = self.x_est[6,0]
		nu_c_temp[1] = self.x_est[7,0]
		nu_c_temp[2] = self.x_est[8,0]
		
		nu_r = nu_arr - np.dot(np.transpose(R),nu_c_temp)

		# derivated (wrt psi) transposed rotation matrix multiplied with current in NED frame
		R_deriv_nu_c = np.zeros((3,1))
		R_deriv_nu_c[0,0] = -np.sin(self.yaw_ang_est)*self.x_est[6,0] + np.cos(self.yaw_ang_est)*self.x_est[7,0]
		R_deriv_nu_c[1,0] = -np.cos(self.yaw_ang_est)*self.x_est[6,0] - np.sin(self.yaw_ang_est)*self.x_est[7,0]

		# df_2/deta
		# N(nu_r) wrt psi
		N_deriv_psi_1 = self.milliampere.damping_wrt_psi(nu_arr, nu_c_temp, self.yaw_ang_est)
		N_deriv_psi_2 = self.milliampere.coriolis_wrt_psi(nu_arr,nu_c_temp,self.yaw_ang_est)
		N_deriv_psi = N_deriv_psi_1 + N_deriv_psi_2

		partial_jacobi_21_1 = -np.dot(self.M_inv,np.dot(N_deriv_psi,nu_r)).reshape(3,1)
		partial_jacobi_21_2 = np.dot(self.M_inv_N,R_deriv_nu_c).reshape(3,1)
		jacobi_21 = np.hstack([np.zeros((3,2)), partial_jacobi_21_1 + partial_jacobi_21_2])

		# df_2/dnu
		# N(nu_r) wrt nu
		N_deriv_u = self.milliampere.coriolis_wrt_u() + self.milliampere.damping_wrt_u(nu_arr, nu_c_temp, self.yaw_ang_est)
		N_deriv_v = self.milliampere.coriolis_wrt_v() + self.milliampere.damping_wrt_v(nu_arr, nu_c_temp, self.yaw_ang_est)
		N_deriv_r = self.milliampere.coriolis_wrt_r() + self.milliampere.damping_wrt_r(nu_arr)

		N_deriv_nu_nu_r = np.hstack([np.dot(N_deriv_u,nu_r).reshape(3,1),np.dot(N_deriv_v,nu_r).reshape(3,1),\
			np.dot(N_deriv_r,nu_r).reshape(3,1)]).astype(float)

		jacobi_22 = -self.M_inv_N - np.matmul(self.M_inv, N_deriv_nu_nu_r)

		# df_2/dV_c
		# N(nu_r) wrt V_c
		N_deriv_vc_x = self.milliampere.coriolis_wrt_vcx(self.yaw_ang_est) + self.milliampere.damping_wrt_vcx(nu_arr, nu_c_temp, self.yaw_ang_est)
		N_deriv_vc_y = self.milliampere.coriolis_wrt_vcy(self.yaw_ang_est) + self.milliampere.damping_wrt_vcy(nu_arr, nu_c_temp, self.yaw_ang_est)
		N_deriv_vc_nu_r = np.hstack([np.dot(N_deriv_vc_x,nu_r).reshape(3,1),np.dot(N_deriv_vc_y,nu_r).reshape(3,1),np.zeros((3,1))]).astype(float)

		v_c_matrix = np.matrix([[1.0,0.0,0.0],[0.0,1.0,0.0],[0.0,0.0,0.0]])
		jacobi_23 = -np.matmul(self.M_inv,N_deriv_vc_nu_r) + np.matmul(self.M_inv_N,np.matmul(np.transpose(R),v_c_matrix))

		jacobi_24 = self.M_inv

		phi_matrix = np.vstack([np.hstack([np.eye(3) + h*jacobi_11, h*jacobi_12, np.zeros((3,3)),np.zeros((3,3))]),\
			np.hstack([h*jacobi_21, np.eye(3) + h*jacobi_22, h*jacobi_23, h*jacobi_24]),\
			np.hstack([np.zeros((3,3)), np.zeros((3,3)), np.eye(3), np.zeros((3,3))]),\
			np.hstack([np.zeros((3,3)), np.zeros((3,3)), np.zeros((3,3)), np.eye(3) -h*(1/self.T_thrust)*np.eye(3)])])

		return phi_matrix

	

	def predict(self,h,tau_input):
		# prediction step
		tau = self.x_est[9:12,0]
		self.x_pred[2,0] = self.x_est[2,0] + h*self.x_est[5,0]
		self.x_pred[2,0] = self.wrap_yaw(self.x_pred[2,0])

		self.x_pred[0,0] = self.x_est[0,0] + h*np.cos(self.yaw_ang_est)*self.x_est[3,0] - h*np.sin(self.yaw_ang_est)*self.x_est[4,0]
		self.x_pred[1,0] = self.x_est[1,0] + h*np.sin(self.yaw_ang_est)*self.x_est[3,0] + h*np.cos(self.yaw_ang_est)*self.x_est[4,0]
		
		# current in BODY frame
		nu_c_rot = np.zeros((3,1))
		nu_c_rot[0,0] = self.x_est[6,0]*np.cos(self.yaw_ang_est) + self.x_est[7,0]*np.sin(self.yaw_ang_est)
		nu_c_rot[1,0] = -self.x_est[6,0]*np.sin(self.yaw_ang_est) + self.x_est[7,0]*np.cos(self.yaw_ang_est)

		# relative velocity in BODY frame
		nu_r = np.zeros((3,))
		nu_r[0] = self.x_est[3,0] - nu_c_rot[0,0]
		nu_r[1] = self.x_est[4,0] - nu_c_rot[1,0]
		nu_r[2] = self.x_est[5,0]

		C = self.milliampere.coriolis(nu_r)
		D = self.milliampere.damping(nu_r)
		N = D
		if self.milliampere.use_coriolis:
			N += C
		self.N = N

		self.M_inv_N = np.matmul(self.M_inv,self.N)
		M_inv_N_nu_r = self.M_inv_N.dot(nu_r)
		M_inv_tau = np.dot(self.M_inv,tau)
		
		self.x_pred[3,0] = self.x_est[3,0] - h*M_inv_N_nu_r[0] + h*M_inv_tau[0]
		self.x_pred[4,0] = self.x_est[4,0] - h*M_inv_N_nu_r[1] + h*M_inv_tau[1]
		self.x_pred[5,0] = self.x_est[5,0] - h*M_inv_N_nu_r[2] + h*M_inv_tau[2]
		self.x_pred[6,0] = self.x_est[6,0]
		self.x_pred[7,0] = self.x_est[7,0]
		self.x_pred[8,0] = self.x_est[8,0]

		self.x_pred[9,0] = self.x_est[9,0] - h*(1/self.T_thrust)*(self.x_est[9,0] - tau_input[0])
		self.x_pred[10,0] = self.x_est[10,0] - h*(1/self.T_thrust)*(self.x_est[10,0] - tau_input[1])
		self.x_pred[11,0] = self.x_est[11,0] - h*(1/self.T_thrust)*(self.x_est[11,0] - tau_input[2])

		Phi = self.updatePhi(h).astype(float)

		Delta = h*self.E.astype(float)

		self.P_pred = np.matmul(Phi,np.matmul(self.P_est,np.transpose(Phi))) \
			+ np.matmul(Delta,np.matmul(self.Q,np.transpose(Delta)))

		# update step
		self.x_prev = self.x_pred
		self.P_prev = self.P_pred


class Caller():
	def __init__(self):
		# read sensors and perform low-pass filtering
		rospy.Subscriber('/guidance/reference/pose', NorthEastHeading, self.callback_eta_ref)
		rospy.Subscriber(topics.nu,SurgeSwayYaw,self.callback_nu)
		rospy.Subscriber(topics.eta,NorthEastHeading,self.callback_eta)
		rospy.Subscriber('/dynamic_positioning/control_action', ThreeDofForce, self.callback_tau)

		self.eta = np.zeros((3,))
		self.nu = np.zeros((3,))
		self.tau = np.zeros((3,))
		self.heading_des = 0.0
		self.Tf = 1/(2*pi*8)
		self.first_eta = True
		self.first_nu = True

		
	def callback_eta(self, data):
		if self.first_eta:
			self.eta = np.array([data.north, data.east, data.heading])
			self.old_time_eta = rospy.Time().now().to_sec()
			self.first_eta = False
		else:
			self.new_time_eta = rospy.Time().now().to_sec()
			h = self.new_time_eta - self.old_time_eta
			eta_new = np.array([data.north, data.east, data.heading])
			a = h/(self.Tf + h)
			self.eta = (1-a)*self.eta + a*eta_new
			self.eta[2] = eta_new[2]
			self.old_time_eta = self.new_time_eta

	def callback_nu(self, data):
		if self.first_nu:
			self.nu = np.array([data.surge, data.sway, data.yaw])
			self.old_time_nu = rospy.Time().now().to_sec()
			self.first_nu = False
		else:
			self.new_time_nu = rospy.Time().now().to_sec()
			nu_new = np.array([data.surge, data.sway, data.yaw])
			h = self.new_time_nu - self.old_time_nu
			a = h/(self.Tf + h)
			self.nu = (1-a)*self.nu + a*nu_new
			self.old_time_nu = self.new_time_nu


	def callback_eta_ref(self,data):
		self.heading_des = data.heading

	def callback_tau(self,data):
		#self.tau = np.zeros((3,)) # used when floating
		self.tau = np.array([[data.X], [data.Y], [data.N]])




rospy.init_node('env_estimation')
mode = ''
def callback_mode(data):
	global mode
	mode = data.data

rospy.Subscriber('/supervisor/mode', String, callback_mode)

x = np.zeros((12,1))
P = np.eye(12)
frequency = 10.00  
Rate = rospy.Rate(frequency) 
kalman = Kalman(x, P)
#scaller = Caller()


if __name__ == "__main__":
	try:
		old_time = rospy.Time().now().to_sec()
		old_time_lp = rospy.Time().now().to_sec()
		# loop continues while ros is not shutdown
		caller = Caller()
		while not rospy.is_shutdown():
			

			if mode in ['joystick','dp']:

				# Estimation
				
				kalman.estimate(caller)

				# time step for prediction
				new_time = rospy.Time().now().to_sec()
				h = new_time - old_time

				# predict the next step
				kalman.predict(h, caller.tau)
				#caller.tau_prev = caller.tau_now
				old_time = new_time
				# Publish the predited and the estimated steps
				kalman.publish_message()
				now = Rate.remaining()
				if now.secs < 0:
					rospy.logwarn('kalman took too long %i %i',now.secs, now.nsecs) 
					# if this is less than zero the node took too long

			Rate.sleep()

	except rospy.ROSInterruptException:
		pass